"""
internal-only classes used to manage information about
source distributions and their contents
"""

import os
import tarfile
import zipfile
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from functools import cached_property
from tempfile import TemporaryDirectory

from ._file_utils import (
    _ArchiveFormat,
    _decompress_zstd_archive,
    _DirectoryInfo,
    _FileInfo,
    _guess_archive_format,
)


@dataclass
class _DistributionSummary:
    archive_format: str
    compressed_size_bytes: int
    directories: list[_DirectoryInfo]
    files: list[_FileInfo]
    original_file: str

    @classmethod
    def from_file(cls, filename: str) -> "_DistributionSummary":
        archive_format = _guess_archive_format(filename)
        compressed_size_bytes = os.path.getsize(filename)
        directories: list[_DirectoryInfo] = []
        files: list[_FileInfo] = []
        if archive_format == _ArchiveFormat.GZIP_TAR:
            with tarfile.open(filename, mode="r:gz") as tf:
                for tar_info in tf.getmembers():
                    if tar_info.isfile():
                        files.append(
                            _FileInfo.from_tarfile_member(
                                archive_file=tf, tar_info=tar_info
                            )
                        )
                    else:
                        directories.append(_DirectoryInfo(name=tar_info.name))
        elif archive_format == _ArchiveFormat.BZIP2_TAR:
            with tarfile.open(filename, mode="r:bz2") as tf:
                for tar_info in tf.getmembers():
                    if tar_info.isfile():
                        files.append(
                            _FileInfo.from_tarfile_member(
                                archive_file=tf, tar_info=tar_info
                            )
                        )
                    else:
                        directories.append(_DirectoryInfo(name=tar_info.name))
        elif archive_format == _ArchiveFormat.CONDA:
            # as of Jan 2023, .conda files are a zip archive containing:
            #   - an uncompressed file 'metadata.json' describing the contents
            #   - 2 zstd-compressed tarfiles with the package contents
            #      - 'info-*.tar.zst'
            #      - 'pkg-*.tar.zst'
            #
            # ref: https://docs.conda.io/projects/conda/en/latest/user-guide/concepts/packages.html#conda-file-format
            with (
                zipfile.ZipFile(filename, mode="r") as f,
                TemporaryDirectory() as tmp_dir,
            ):
                for zip_info in f.infolist():
                    # case 1 - is a directory
                    if zip_info.is_dir():
                        directories.append(_DirectoryInfo(name=zip_info.filename))
                    # case 2 - is a file but not one of the zstandard-compressed ones
                    elif not zip_info.filename.lower().endswith("tar.zst"):
                        files.append(
                            _FileInfo.from_zipfile_member(
                                archive_file=f, zip_info=zip_info
                            )
                        )
                    # case 3 - one of the zstandard-compressed archives
                    else:
                        full_path = os.path.join(tmp_dir, zip_info.filename)
                        # ref: https://stackoverflow.com/a/55260983/3986677
                        #
                        # decompress and write to a regular tarfile
                        with zipfile.ZipFile(filename, mode="r") as zf:
                            zf.extractall(path=tmp_dir, members=[zip_info.filename])

                            # decompress the .tar.zst to just .tar
                            decompressed_tar_path = full_path.lower().replace(
                                ".tar.zst", ".tar"
                            )
                            _decompress_zstd_archive(
                                tar_zst_file=full_path,
                                decompressed_tar_path=decompressed_tar_path,
                            )

                        # do tarfile things
                        with tarfile.open(decompressed_tar_path, mode="r") as tf:
                            for tar_info in tf.getmembers():
                                if tar_info.isfile():
                                    files.append(
                                        _FileInfo.from_tarfile_member(
                                            archive_file=tf, tar_info=tar_info
                                        )
                                    )
                                else:
                                    directories.append(
                                        _DirectoryInfo(name=tar_info.name)
                                    )
        elif archive_format == _ArchiveFormat.ZIP:
            # assume anything else can be opened with zipfile
            with zipfile.ZipFile(filename, mode="r") as f:
                for zip_info in f.infolist():
                    if not zip_info.is_dir():
                        files.append(
                            _FileInfo.from_zipfile_member(
                                archive_file=f, zip_info=zip_info
                            )
                        )
                    else:
                        directories.append(_DirectoryInfo(name=zip_info.filename))

        return cls(
            archive_format=archive_format,
            compressed_size_bytes=compressed_size_bytes,
            directories=directories,
            files=files,
            original_file=filename,
        )

    @property
    def all_paths(self) -> list[str]:
        return self.file_paths + self.directory_paths

    @property
    def compiled_objects(self) -> list[_FileInfo]:
        return [f for f in self.files if f.is_compiled is True]

    @cached_property
    def count_by_file_extension(self) -> dict[str, int]:
        return {
            file_extension: len(file_list)
            for file_extension, file_list in self.files_by_extension.items()
        }

    @property
    def directory_paths(self) -> list[str]:
        return [d.name for d in self.directories]

    @cached_property
    def files_by_extension(self) -> dict[str, list[_FileInfo]]:
        out: dict[str, list[_FileInfo]] = defaultdict(list)
        for f in self.files:
            out[f.file_extension].append(f)
        return out

    @property
    def file_paths(self) -> list[str]:
        return [f.name for f in self.files]

    @property
    def num_directories(self) -> int:
        return len(self.directories)

    @property
    def num_files(self) -> int:
        return len(self.files)

    def get_largest_files(self, n: int) -> list[_FileInfo]:
        return sorted(
            self.files, key=lambda f: f.uncompressed_size_bytes, reverse=True
        )[:n]

    @property
    def uncompressed_size_bytes(self) -> int:
        return sum(f.uncompressed_size_bytes for f in self.files)

    @property
    def size_by_file_extension(self) -> "OrderedDict[str, int]":
        """
        Aggregate file sizes in a distribution by extension.

        :return: An OrderedDict where keys are file extensions and values are the total size in
                 bytes (uncompressed) occupied by such files in the distribution.
                 Sorted in descending order by size.
        """
        summary_dict = {
            file_extension: sum(f.uncompressed_size_bytes for f in files)
            for file_extension, files in self.files_by_extension.items()
        }
        sorted_sizes = list(summary_dict.items())
        sorted_sizes.sort(key=lambda x: x[1], reverse=True)
        out = OrderedDict()
        for file_extension, size_in_bytes in sorted_sizes:
            out[file_extension] = size_in_bytes
        return out



"""
Central location for weird import stuff used to make the project compatible
with a wide range of dependency versions.
"""

from typing import Any

try:  # pragma: no cover
    import tomllib
except ModuleNotFoundError:  # pragma: no cover
    import tomli as tomllib  # type: ignore[no-redef]


def _import_zstandard() -> Any:
    try:
        import zstandard

        return zstandard
    except ModuleNotFoundError as err:
        err_msg = (
            "Checking zstd-compressed files requires the 'zstandard' library. "
            "Install it with e.g. 'pip install zstandard' or 'conda install conda-forge::zstandard'."
        )
        raise ModuleNotFoundError(err_msg) from err


__all__ = ["_import_zstandard", "tomllib"]



"""
Manages configuration of ``pydistcheck` CLI, including:

  * validating configuration values
  * updating configuuration from files
"""

import os
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

from ._compat import tomllib

# putting this in a module-level set to save on the cost of re-computing it inside methods
# in `_Config` that validate configuration.
#
# unit tests confirm that it matches the `_Config` class, so it shouldn't ever drift from that class
_ALLOWED_CONFIG_VALUES = {
    "expected_directories",
    "expected_files",
    "ignore",
    "inspect",
    "max_allowed_files",
    "max_allowed_size_compressed",
    "max_allowed_size_uncompressed",
    "max_path_length",
    "output_file_size_precision",
    "output_file_size_unit",
    "select",
}

_EXPECTED_DIRECTORIES = (
    "!*/.appveyor",
    "!*/.binder",
    "!*/.circleci",
    "!*/.git",
    "!*/.github",
    "!*/.idea",
    "!*/.pytest_cache",
    "!*/.mypy_cache",
)

_EXPECTED_FILES = (
    "!*/appveyor.yml",
    "!*/.appveyor.yml",
    "!*/azure-pipelines.yml",
    "!*/.azure-pipelines.yml",
    "!*/.cirrus.star",
    "!*/.cirrus.yml",
    "!*/codecov.yml",
    "!*/.codecov.yml",
    "!*/.DS_Store",
    "!*/.gitignore",
    "!*/.gitpod.yml",
    "!*/.hadolint.yaml",
    "!*/.lycheecache",
    "!*/.lycheeignore",
    "!*/.readthedocs.yaml",
    "!*/.travis.yml",
    "!*/vsts-ci.yml",
    "!*/.vsts-ci.yml",
)


@dataclass
class _Config:
    expected_directories: Sequence[str] = _EXPECTED_DIRECTORIES
    expected_files: Sequence[str] = _EXPECTED_FILES
    ignore: Sequence[str] = ()
    inspect: bool = False
    max_allowed_files: int = 2000
    max_allowed_size_compressed: str = "50M"
    max_allowed_size_uncompressed: str = "75M"
    max_path_length: int = 200
    output_file_size_precision: int = 3
    output_file_size_unit: str = "auto"
    select: Sequence[str] = ()

    def __setattr__(self, name: str, value: Any) -> None:
        attr_name = name.replace("-", "_")
        if attr_name not in _ALLOWED_CONFIG_VALUES:
            raise ValueError(
                f"Configuration value '{name}' is not recognized by pydistcheck"
            )
        object.__setattr__(self, attr_name, value)

    def update_from_dict(self, input_dict: dict[str, Any]) -> "_Config":
        for k, v in input_dict.items():
            setattr(self, k, v)
        return self

    def update_from_toml(self, toml_file: str) -> "_Config":
        if not os.path.exists(toml_file):
            return self

        tool_options: dict[str, Any] = {}
        with open(toml_file, "rb") as f:
            config_dict = tomllib.load(f)
            tool_options = config_dict.get("tool", {}).get("pydistcheck", {})

        self.update_from_dict(tool_options)
        return self



import os
import uuid
from copy import deepcopy
from dataclasses import fields

import pytest

from pydistcheck._config import _ALLOWED_CONFIG_VALUES, _Config


@pytest.fixture(scope="function")
def base_config():
    return _Config(
        inspect=False,
        max_allowed_files=7,
        max_allowed_size_compressed="1G",
        max_allowed_size_uncompressed="18K",
    )


def test_allowed_config_values_constant_matches_class():
    fields_from_class = {f.name for f in fields(_Config)}
    assert fields_from_class == _ALLOWED_CONFIG_VALUES


def test_config_fixture_works_as_expected(base_config):
    assert base_config.inspect is False
    assert base_config.max_allowed_files == 7
    assert base_config.max_allowed_size_compressed == "1G"
    assert base_config.max_allowed_size_uncompressed == "18K"


def test_config_can_be_initialized_without_any_arguments():
    config = _Config()
    assert config.inspect is False
    assert config.max_allowed_files == 2000
    assert config.max_allowed_size_compressed == "50M"
    assert config.max_allowed_size_uncompressed == "75M"


def test_update_from_dict_raises_exception_for_first_bad_value_encountered(base_config):
    with pytest.raises(
        ValueError, match=r"Configuration value 'a' is not recognized by pydistcheck"
    ):
        base_config.update_from_dict({"a": 7, "b": 8, "inspect": False})


def test_update_from_dict_works_even_if_dict_does_not_include_all_config_values(
    base_config,
):
    assert base_config.inspect is False
    base_config.update_from_dict({"inspect": True})
    assert base_config.inspect is True


def test_update_from_dict_works_when_changing_all_values(base_config):
    assert base_config.ignore == ()
    assert base_config.inspect is False
    assert base_config.max_allowed_files == 7
    assert base_config.max_allowed_size_compressed == "1G"
    assert base_config.max_allowed_size_uncompressed == "18K"
    assert base_config.output_file_size_precision == 3
    assert base_config.output_file_size_unit == "auto"
    assert base_config.select == ()
    patch_dict = {
        "expected_directories": "!*/tests",
        "expected_files": "!*.xlsx,!data/*.csv",
        "ignore": ["path-contains-spaces", "too-many-files"],
        "inspect": True,
        "max_allowed_files": 8,
        "max_allowed_size_compressed": "2G",
        "max_allowed_size_uncompressed": "141K",
        "max_path_length": 600,
        "output_file_size_precision": 2,
        "output_file_size_unit": "GB",
        "select": ["distro-too-large-compressed"],
    }
    assert set(patch_dict.keys()) == _ALLOWED_CONFIG_VALUES, (
        "this test needs to be updated"
    )
    base_config.update_from_dict(patch_dict)
    assert base_config.expected_directories == "!*/tests"
    assert base_config.expected_files == "!*.xlsx,!data/*.csv"
    assert base_config.ignore == ["path-contains-spaces", "too-many-files"]
    assert base_config.inspect is True
    assert base_config.max_allowed_files == 8
    assert base_config.max_allowed_size_compressed == "2G"
    assert base_config.max_allowed_size_uncompressed == "141K"
    assert base_config.max_path_length == 600
    assert base_config.output_file_size_precision == 2
    assert base_config.output_file_size_unit == "GB"
    assert base_config.select == ["distro-too-large-compressed"]


def test_update_from_toml_silently_returns_self_if_file_does_not_exist(base_config):
    original_config = deepcopy(base_config)
    out = base_config.update_from_toml(toml_file=f"{uuid.uuid4().hex}.toml")
    assert out == original_config


def test_update_from_toml_works_for_files_with_no_pydistcheck_configuration(
    base_config, tmpdir
):
    original_config = deepcopy(base_config)
    temp_file = os.path.join(tmpdir, f"{uuid.uuid4().hex}.toml")
    with open(temp_file, "w") as f:
        f.write("""[tool.pylint]\n""")
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config == original_config


def test_update_from_toml_works_for_files_with_empty_pydistcheck_configuration(
    base_config, tmpdir
):
    original_config = deepcopy(base_config)
    temp_file = os.path.join(tmpdir, f"{uuid.uuid4().hex}.toml")
    with open(temp_file, "w") as f:
        f.write("""\n[tool.pylint]\n[tool.pydistcheck]\n""")
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config == original_config


@pytest.mark.parametrize(
    "config_key", ["max_allowed_size_compressed", "max-allowed-size_compressed"]
)
def test_update_from_toml_works_with_underscores_and_hyphens(
    base_config, tmpdir, config_key
):
    temp_file = os.path.join(tmpdir, f"{uuid.uuid4().hex}.toml")
    with open(temp_file, "w") as f:
        f.write(f"[tool.pylint]\n[tool.pydistcheck]\n{config_key} = '2.5G'\n")
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config.max_allowed_size_compressed == "2.5G"


@pytest.mark.parametrize("use_hyphens", [True, False])
def test_update_from_toml_works_with_all_config_values(
    base_config, tmpdir, use_hyphens
):
    temp_file = os.path.join(tmpdir, f"{uuid.uuid4().hex}.toml")
    patch_dict = {
        "expected_directories": "[\n'!tests/*'\n]",
        "expected_files": "[\n'!*.pq',\n'!*/tests/data/*.csv']",
        "ignore": "[\n'path-contains-spaces',\n'too-many-files'\n]",
        "inspect": "true",
        "max_allowed_files": 8,
        "max_allowed_size_compressed": "'3G'",
        "max_allowed_size_uncompressed": "'4.12G'",
        "max_path_length": 25,
        "output_file_size_precision": 2,
        "output_file_size_unit": "'Mi'",
        "select": "[\n'mixed-file-extensions',\n'path-contains-non-ascii-characters'\n]",
    }
    assert set(patch_dict.keys()) == _ALLOWED_CONFIG_VALUES, (
        "this test needs to be updated"
    )
    if use_hyphens:
        patch_dict = {k.replace("_", "-"): v for k, v in patch_dict.items()}
    with open(temp_file, "w") as f:
        lines = ["[tool.pylint]", "[tool.pydistcheck]"] + [
            f"{k} = {v}" for k, v in patch_dict.items()
        ]
        f.write("\n".join(lines))
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config.expected_directories == ["!tests/*"]
    assert base_config.expected_files == ["!*.pq", "!*/tests/data/*.csv"]
    assert base_config.ignore == ["path-contains-spaces", "too-many-files"]
    assert base_config.inspect is True
    assert base_config.max_allowed_files == 8
    assert base_config.max_allowed_size_compressed == "3G"
    assert base_config.max_allowed_size_uncompressed == "4.12G"
    assert base_config.max_path_length == 25
    assert base_config.output_file_size_precision == 2
    assert base_config.output_file_size_unit == "Mi"
    assert base_config.select == [
        "mixed-file-extensions",
        "path-contains-non-ascii-characters",
    ]



import pytest

from pydistcheck._utils import _FileSize


@pytest.mark.parametrize(
    ("unit_str", "expected_total_bytes"),
    [
        ("B", 3),
        ("K", 3 * 1024),
        ("KB", 3e3),
        ("kB", 3e3),
        ("kb", 3e3),
        ("Ki", 3 * 1024),
        ("ki", 3 * 1024),
        ("M", 3 * (1024**2)),
        ("MB", 3e6),
        ("mB", 3e6),
        ("mb", 3e6),
        ("Mi", 3 * (1024**2)),
        ("mi", 3 * (1024**2)),
        ("G", 3 * (1024**3)),
        ("GB", 3e9),
        ("gB", 3e9),
        ("gb", 3e9),
        ("Gi", 3 * (1024**3)),
        ("gi", 3 * (1024**3)),
    ],
)
def test_file_size_minimally_works(expected_total_bytes, unit_str):
    fs = _FileSize(num=3.0, unit_str=unit_str)
    assert fs.total_size_bytes == int(expected_total_bytes)


def test_file_size_comparisons_work():
    fs_5mb = _FileSize(num=5.6, unit_str="M")
    fs_6mb = _FileSize(num=4 * (1024**3), unit_str="B")

    assert fs_5mb == fs_5mb  # noqa: PLR0124
    assert fs_5mb != fs_6mb
    assert fs_5mb != fs_6mb
    assert fs_5mb < fs_6mb
    assert fs_5mb <= fs_6mb
    assert fs_6mb >= fs_5mb
    assert fs_6mb > fs_5mb


@pytest.mark.parametrize(
    "file_size",
    [
        _FileSize.from_number(3 * 1024**2),
        _FileSize.from_string("3M"),
        _FileSize.from_string("3.0M"),
        _FileSize.from_string("3.0000M"),
    ],
)
def test_file_size_from_different_inputs_all_parsed_consistently(file_size):
    assert file_size == _FileSize(num=3.0, unit_str="M")


def test_file_size_from_string_works():
    # 1 letter
    assert _FileSize.from_string("57B").total_size_bytes == 57
    assert _FileSize.from_string("57K").total_size_bytes == 57 * 1024
    assert _FileSize.from_string("57M").total_size_bytes == 57 * 1024**2
    assert _FileSize.from_string("57G").total_size_bytes == 57 * 1024**3

    # 2 letter
    assert _FileSize.from_string("57KB").total_size_bytes == 57 * 1e3
    assert _FileSize.from_string("57MB").total_size_bytes == 57 * 1e6
    assert _FileSize.from_string("57GB").total_size_bytes == 57 * 1e9

    # decimals
    assert _FileSize.from_string("0.005GB").total_size_bytes == 0.005 * 1e9
    assert _FileSize.from_string("0.05GB").total_size_bytes == 0.05 * 1e9
    assert _FileSize.from_string("0.5GB").total_size_bytes == 0.5 * 1e9
    assert _FileSize.from_string("5.0GB").total_size_bytes == 5 * 1e9
    assert _FileSize.from_string("5.000GB").total_size_bytes == 5 * 1e9
    assert _FileSize.from_string("5.1GB").total_size_bytes == 5.1 * 1e9
    assert _FileSize.from_string("5.17GB").total_size_bytes == 5.17 * 1e9
    assert _FileSize.from_string("5.234GB").total_size_bytes == 5.234 * 1e9


def test_file_size_from_number_switches_unit_str_based_on_size():
    assert _FileSize.from_number(1.1) == _FileSize(num=1.1, unit_str="B")
    # fractional bytes don't make sense here, so some rounding happens
    # e.g., 0.1 KB is technically 102.4 bytes, which gets rounded to 102
    assert _FileSize.from_number(102) == _FileSize(num=0.1, unit_str="K")
    # this value is 3.456789 * 10**3 = 3711698926.043136
    assert _FileSize.from_number(3711698926) == _FileSize(num=3.456789, unit_str="G")


@pytest.mark.parametrize(
    ("precision", "unit_str", "expected_str"),
    [
        # 1 significant digit
        (1, "B", "1234567.0B"),
        (2, "B", "1234567.0B"),
        (4, "B", "1234567.0B"),
        (7, "B", "1234567.0B"),
        # 3 significant digits
        (1, "KB", "1234.6KB"),
        (2, "KB", "1234.57KB"),
        (4, "KB", "1234.567KB"),
        (7, "KB", "1234.567KB"),
        # 9 significant digits
        (1, "GB", "0.0GB"),
        (2, "GB", "0.0GB"),
        (4, "GB", "0.0012GB"),
        (7, "GB", "0.0012346GB"),
        (9, "GB", "0.001234567GB"),
    ],
)
def test_file_size_to_string_works(precision, unit_str, expected_str):
    fs = _FileSize(num=1234567, unit_str="B")
    assert fs.to_string(precision=precision, unit_str=unit_str) == expected_str


@pytest.mark.parametrize(
    ("file_size", "expected_str"),
    [
        (_FileSize.from_number(110), "0.107K"),
        (_FileSize.from_number(150), "0.146K"),
        (_FileSize.from_number(1024), "1.0K"),
        (_FileSize.from_number(100 * 1024), "100.0K"),
        (_FileSize.from_number(200 * 1024), "0.195M"),
        (_FileSize.from_number(1024**2), "1.0M"),
        (_FileSize.from_number(400 * 1024**2), "0.391G"),
        (_FileSize.from_number(405 * 1024**2), "0.396G"),
        (_FileSize.from_number(1024**3), "1.0G"),
        (_FileSize.from_number(int(7.5 * 1024**3)), "7.5G"),
    ],
)
def test_file_size_string_representation_looks_correct(file_size, expected_str):
    assert str(file_size) == expected_str



"""
functions used to analyze compiled objects
"""

import re
import subprocess

_COMMAND_FAILED = "__command_failed__"
_NO_DEBUG_SYMBOLS = "__no_debug_symbols_found__"
_TOOL_NOT_AVAILABLE = "__tool_not_available__"

# This specific debug symbol is added by Apple's "strip" on
# Mach-O files and can safely be ignored for the purpose of detecting debug symbols. See:
# https://github.com/jameslamb/pydistcheck/issues/235
_MACHO_STRIP_SYMBOL = "radr://5614542"


def _run_command(args: list[str]) -> str:
    try:
        stdout = subprocess.run(args, capture_output=True, check=True).stdout
        # Use latin1 encoding, which can handle any byte value without data loss.
        # See https://github.com/jameslamb/pydistcheck/issues/206 for rationale.
        return stdout.decode("latin1")
    except subprocess.CalledProcessError:
        return _COMMAND_FAILED
    except FileNotFoundError:
        return _TOOL_NOT_AVAILABLE


# commands to dump symbol information, and regular expressions which, if they match
# any lines in the output, indicate that debug symbols have been found
# fmt: off
_COMMANDS_TO_PATTERNS = [
    (["dsymutil", "-s"],                           r"\(N_OSO[\t ]+\)"),
    (["objdump", "--all-headers"],                 r"[\t ]+\.debug_line[\t ]+"),
    (["objdump", "--macho", "--all-headers"],      r"[\t ]+\.debug_line[\t ]+"),
    (["objdump", "-W"],                            r"^Contents of the \.debug"),
    (["objdump", "--macho", "-W"],                 r"^Contents of the \.debug"),
    (["objdump", "-g"],                            r"^Contents of the \.debug"),
    (["objdump", "--macho", "-g"],                 r"^Contents of the \.debug"),
    (["llvm-objdump", "--all-headers"],            r"[\t ]+\.debug_line[\t ]+"),
    (["llvm-objdump", "--macho", "--all-headers"], r"[\t ]+\.debug_line[\t ]+"),
    (["llvm-objdump", "-W"],                       r"^Contents of the \.debug"),
    (["llvm-objdump", "--macho", "-W"],            r"^Contents of the \.debug"),
    (["llvm-objdump", "-g"],                       r"^Contents of the \.debug"),
    (["llvm-objdump", "--macho", "-g"],            r"^Contents of the \.debug"),
    (["readelf", "-S"],                            r"[\t ]+\.debug_[a-z]+[\t ]+")
]
# fmt: on


def _look_for_debug_symbols(lib_file: str) -> tuple[bool, str]:
    for cmd_args, pattern in _COMMANDS_TO_PATTERNS:
        stdout = _run_command(args=[*cmd_args, lib_file])
        contains_debug_symbols = any(
            bool(re.search(pattern, x)) for x in stdout.split("\n")
        )
        if contains_debug_symbols:
            return True, " ".join(cmd_args)
    # if you get here, no debug symbols were found by any tools
    return False, _NO_DEBUG_SYMBOLS


def _get_symbols(cmd_args: list[str], lib_file: str) -> str:
    syms = _run_command(args=[*cmd_args, lib_file])
    return "\n".join(
        [line for line in syms.split("\n") if line and _MACHO_STRIP_SYMBOL not in line]
    )


def _nm_reports_debug_symbols(tool_name: str, lib_file: str) -> tuple[bool, str]:
    exported_symbols = _get_symbols(cmd_args=[tool_name], lib_file=lib_file)
    all_symbols = _get_symbols(cmd_args=[tool_name, "-a"], lib_file=lib_file)
    return exported_symbols != all_symbols, f"{tool_name} -a"


def _file_has_debug_symbols(file_absolute_path: str) -> tuple[bool, str]:
    # test with tools that produce debug symbols that can be matched with a regex
    has_debug_symbols, cmd_str = _look_for_debug_symbols(lib_file=file_absolute_path)
    if has_debug_symbols:
        return True, cmd_str

    # "nm"'s test is like "check if the output is different when a flag is supplied",
    # instead of "test if this tool produces output matching this regex",
    # so it runs separately here
    for nm_tool in ["nm", "llvm-nm"]:
        has_debug_symbols, cmd_str = _nm_reports_debug_symbols(
            tool_name=nm_tool,
            lib_file=file_absolute_path,
        )
        if has_debug_symbols:  # pragma: no cover
            return True, cmd_str

    # at this point, none of the checks found debug symbols
    return False, cmd_str



"""
CLI entrypoints
"""

import sys
from collections.abc import Sequence

import click

from pydistcheck import __version__ as _VERSION

from ._checks import (
    ALL_CHECKS,
    _CompiledObjectsDebugSymbolCheck,
    _DistroTooLargeCompressedCheck,
    _DistroTooLargeUnCompressedCheck,
    _ExpectedFilesCheck,
    _FileCountCheck,
    _FilesOnlyDifferByCaseCheck,
    _MixedFileExtensionCheck,
    _NonAsciiCharacterCheck,
    _PathTooLongCheck,
    _SpacesInPathCheck,
    _UnexpectedFilesCheck,
)
from ._config import _Config
from ._distribution_summary import _DistributionSummary
from ._inspect import inspect_distribution
from ._utils import _FileSize


class ExitCodes:
    OK = 0
    CHECK_ERRORS = 1
    UNSUPPORTED_FILE_TYPE = 2


@click.command()
@click.argument(
    "filepaths",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--version",
    is_flag=True,
    show_default=False,
    default=False,
    help="Print the version of pydistcheck and exit.",
)
@click.option(
    "--config",
    type=click.Path(exists=True),
    default=None,
    help=(
        "Path to a TOML file containing a [tool.pydistcheck] table. "
        "If provided, pyproject.toml will be ignored."
    ),
)
@click.option(
    "--ignore",
    multiple=True,
    default=_Config.ignore,
    help=(
        "ID of a check to skip, e.g. 'compiled-objects-have-debug-symbols'. "
        "See https://pydistcheck.readthedocs.io/en/docs-fix/check-reference.html for a "
        "complete list of valid options. Can be passed multiple times."
    ),
)
@click.option(
    "--select",
    multiple=True,
    default=_Config.select,
    help=(
        "ID of a check to include, e.g. 'compiled-objects-have-debug-symbols'. "
        "If this is provided even once, pydistcheck will only run the checks specified this way. "
        "See https://pydistcheck.readthedocs.io/en/docs-fix/check-reference.html for a "
        "complete list of valid options. Can be passed multiple times."
    ),
)
@click.option(
    "--inspect",
    is_flag=True,
    show_default=False,
    default=_Config.inspect,
    help=(
        "Print a summary of the distribution, like its total size and largest files."
    ),
)
@click.option(
    "--expected-directories",
    multiple=True,
    default=_Config.expected_directories,
    show_default=True,
    help=(
        "Pattern matching directories that are expected to be found in the distribution. "
        "Prefix with '!' to indicate a pattern which should NOT match any of the distribution's "
        "contents. Other than that possible leading '!', patterns should be in the format understood by "
        "``fnmatch.fnmatchcase()`` (https://docs.python.org/3/library/fnmatch.html). "
        "Can be passed multiple times."
    ),
)
@click.option(
    "--expected-files",
    multiple=True,
    default=_Config.expected_files,
    show_default=True,
    help=(
        "Pattern matching files that are expected to be found in the distribution. "
        "Prefix with '!' to indicate a pattern which should NOT match any of the distribution's "
        "contents. Other than that possible leading '!', patterns should be in the format understood by "
        "``fnmatch.fnmatchcase()`` (https://docs.python.org/3/library/fnmatch.html). "
        "Can be passed multiple times."
    ),
)
@click.option(
    "--max-allowed-files",
    default=_Config.max_allowed_files,
    show_default=True,
    type=int,
    help="maximum number of files allowed in the distribution",
)
@click.option(
    "--max-allowed-size-compressed",
    default=_Config.max_allowed_size_compressed,
    show_default=True,
    type=str,
    help=(
        "maximum allowed compressed size, a string like '1.5MB' indicating"
        " '1.5 megabytes'. Supported units:\n"
        "  - B = bytes\n"
        "  - KB = kilobytes\n"
        "  - K, Ki = kibibytes\n"
        "  - MB = megabytes\n"
        "  - M, Mi = mebibytes\n"
        "  - GB = gigabytes\n"
        "  - G, Gi = gibibytes"
    ),
)
@click.option(
    "--max-allowed-size-uncompressed",
    default=_Config.max_allowed_size_uncompressed,
    show_default=True,
    type=str,
    help=(
        "maximum allowed uncompressed size, a string like '1.5MB' indicating"
        " '1.5 megabytes'. Supported units:\n"
        "  - B = bytes\n"
        "  - KB = kilobytes\n"
        "  - K, Ki = kibibytes\n"
        "  - MB = megabytes\n"
        "  - M, Mi = mebibytes\n"
        "  - GB = gigabytes\n"
        "  - G, Gi = gibibytes"
    ),
)
@click.option(
    "--max-path-length",
    default=_Config.max_path_length,
    show_default=True,
    type=int,
    help="Maximum allowed filepath length for files or directories in the distribution.",
)
@click.option(
    "--output-file-size-precision",
    default=_Config.output_file_size_precision,
    show_default=True,
    type=int,
    help=(
        "Number of significant digits to use when rounding file sizes in printed output."
    ),
)
@click.option(
    "--output-file-size-unit",
    default=_Config.output_file_size_unit,
    show_default=True,
    type=str,
    help=(
        "Unit for all file sizes reported in outputs (e.g. log messages, --inspect output)"
        " Supported units:\n"
        "  - B = bytes\n"
        "  - KB = kilobytes\n"
        "  - K, Ki = kibibytes\n"
        "  - MB = megabytes\n"
        "  - M, Mi = mebibytes\n"
        "  - GB = gigabytes\n"
        "  - G, Gi = gibibytes"
    ),
)
def check(  # noqa: PLR0913
    *,
    filepaths: str,
    version: bool,
    config: str,
    expected_directories: Sequence[str],
    expected_files: Sequence[str],
    ignore: Sequence[str],
    inspect: bool,
    max_allowed_files: int,
    max_allowed_size_compressed: str,
    max_allowed_size_uncompressed: str,
    max_path_length: int,
    output_file_size_precision: int,
    output_file_size_unit: str,
    select: Sequence[str],
) -> None:
    """
    Run the contents of a distribution through a set of checks, and warn about
    any problematic characteristics that are detected.

    Exit codes:

      0 = no issues detected\n
      1 = issues detected
    """
    if version:
        print(f"pydistcheck {_VERSION}")
        sys.exit(ExitCodes.OK)

    print("==================== running pydistcheck ====================")
    filepaths_to_check = [click.format_filename(f) for f in filepaths]
    conf = _Config()
    kwargs = {
        "ignore": ignore,
        "inspect": inspect,
        "max_allowed_files": max_allowed_files,
        "max_allowed_size_compressed": max_allowed_size_compressed,
        "max_allowed_size_uncompressed": max_allowed_size_uncompressed,
        "max_path_length": max_path_length,
        "output_file_size_precision": output_file_size_precision,
        "output_file_size_unit": output_file_size_unit,
        "select": select,
        "expected_directories": expected_directories,
        "expected_files": expected_files,
    }
    kwargs_that_differ_from_defaults = {
        k: v for k, v in kwargs.items() if v != getattr(conf, k)
    }
    if config is not None:
        conf.update_from_toml(toml_file=click.format_filename(config))
    else:
        conf.update_from_toml(toml_file="pyproject.toml")
    conf.update_from_dict(input_dict=kwargs_that_differ_from_defaults)

    checks_to_ignore = {x for x in conf.ignore if x.strip()}
    unrecognized_checks = checks_to_ignore - ALL_CHECKS
    if unrecognized_checks:
        # converting to list + sorting here so outputs are deterministic
        # (since sets don't guarantee ordering)
        error_str = ",".join(sorted(unrecognized_checks))
        print(
            f"ERROR: found the following unrecognized checks passed via '--ignore': {error_str}"
        )
        sys.exit(1)

    selected_checks = {x for x in conf.select if x.strip()}
    unrecognized_checks = selected_checks - ALL_CHECKS
    if unrecognized_checks:
        # converting to list + sorting here so outputs are deterministic
        # (since sets don't guarantee ordering)
        error_str = ",".join(sorted(unrecognized_checks))
        print(
            f"ERROR: found the following unrecognized checks passed via '--select': {error_str}"
        )
        sys.exit(1)

    checks = [
        _CompiledObjectsDebugSymbolCheck(),
        _DistroTooLargeCompressedCheck(
            max_allowed_size_bytes=_FileSize.from_string(
                size_str=conf.max_allowed_size_compressed
            ).total_size_bytes,
            output_file_size_precision=conf.output_file_size_precision,
            output_file_size_unit=conf.output_file_size_unit,
        ),
        _DistroTooLargeUnCompressedCheck(
            max_allowed_size_bytes=_FileSize.from_string(
                size_str=conf.max_allowed_size_uncompressed
            ).total_size_bytes,
            output_file_size_precision=conf.output_file_size_precision,
            output_file_size_unit=conf.output_file_size_unit,
        ),
        _ExpectedFilesCheck(
            directory_patterns=expected_directories,
            file_patterns=expected_files,
        ),
        _FileCountCheck(max_allowed_files=conf.max_allowed_files),
        _FilesOnlyDifferByCaseCheck(),
        _MixedFileExtensionCheck(),
        _PathTooLongCheck(max_path_length=conf.max_path_length),
        _SpacesInPathCheck(),
        _UnexpectedFilesCheck(
            directory_patterns=expected_directories,
            file_patterns=expected_files,
        ),
        _NonAsciiCharacterCheck(),
    ]

    # if 'select' is non-empty, use only the checks indicated by that option
    if selected_checks:
        checks = [c for c in checks if c.check_name in selected_checks]
    # otherwise, run all checks except those indicated by 'ignore'
    elif checks_to_ignore:
        checks = [c for c in checks if c.check_name not in checks_to_ignore]

    any_errors_found = False
    for filepath in filepaths_to_check:
        print(f"\nchecking '{filepath}'")

        try:
            summary = _DistributionSummary.from_file(filename=filepath)
        except ValueError as err:
            print(f"error: {err}")
            sys.exit(ExitCodes.UNSUPPORTED_FILE_TYPE)

        if conf.inspect:
            print("----- package inspection summary -----")
            inspect_distribution(
                summary=summary,
                config=conf,
            )

        print("------------ check results -----------")
        errors: list[str] = []
        for this_check in checks:
            errors += this_check(distro_summary=summary)

        for i, error_msg in enumerate(sorted(errors)):
            print(f"{i + 1}. {error_msg}")

        num_errors_for_this_file = len(errors)
        if num_errors_for_this_file:
            any_errors_found = True

        print(f"errors found while checking: {num_errors_for_this_file}")

    print("\n==================== done running pydistcheck ===============")

    # now that all files have been checked, be sure to exit with a non-0 code
    # if any errors were found
    if any_errors_found:
        sys.exit(ExitCodes.CHECK_ERRORS)
    else:
        sys.exit(ExitCodes.OK)



import os
import zipfile

import pytest

from pydistcheck._distribution_summary import _DistributionSummary, _FileInfo

BASE_PACKAGE_SDISTS = ["base-package-0.1.0.tar.gz", "base-package-0.1.0.zip"]
MACOS_SUFFIX = "macosx_10_15_x86_64.macosx_11_6_x86_64.macosx_12_5_x86_64.whl"
MANYLINUX_SUFFIX = "manylinux_2_28_x86_64.manylinux_2_5_x86_64.manylinux1_x86_64.whl"
BASE_WHEELS = [
    f"baseballmetrics-0.1.0-py3-none-{MACOS_SUFFIX}",
    f"baseballmetrics-0.1.0-py3-none-{MANYLINUX_SUFFIX}",
]
WINDOWS_WHEELS = ["lightgbm-3.3.5-py3-none-win_amd64.whl"]
TEST_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


@pytest.mark.parametrize("distro_file", BASE_PACKAGE_SDISTS)
def test_distribution_summary_basically_works(distro_file):
    ds = _DistributionSummary.from_file(os.path.join(TEST_DATA_DIR, distro_file))

    # should correctly capture the contents:
    #   * 3 directories
    #   * 11 files
    assert len(ds.files) == 11
    assert ds.num_files == 11
    assert len(ds.directories) == 3
    assert ds.num_directories == 3

    # file_paths should include all the files and directories
    # (NOTE: zip adds trailing slashes to directories, tar does not)
    if distro_file.endswith("zip"):
        expected_dir_paths = [
            "base-package-0.1.0/",
            "base-package-0.1.0/base_package/",
            "base-package-0.1.0/base_package.egg-info/",
        ]
    else:
        expected_dir_paths = [
            "base-package-0.1.0",
            "base-package-0.1.0/base_package",
            "base-package-0.1.0/base_package.egg-info",
        ]
    expected_file_paths = [
        "base-package-0.1.0/setup.cfg",
        "base-package-0.1.0/PKG-INFO",
        "base-package-0.1.0/setup.py",
        "base-package-0.1.0/LICENSE.txt",
        "base-package-0.1.0/README.md",
        "base-package-0.1.0/base_package/__init__.py",
        "base-package-0.1.0/base_package/thing.py",
        "base-package-0.1.0/base_package.egg-info/top_level.txt",
        "base-package-0.1.0/base_package.egg-info/dependency_links.txt",
        "base-package-0.1.0/base_package.egg-info/SOURCES.txt",
        "base-package-0.1.0/base_package.egg-info/PKG-INFO",
    ]
    expected_all_paths = expected_dir_paths + expected_file_paths

    assert sorted(ds.all_paths) == sorted(expected_all_paths)
    assert ds.compiled_objects == []
    assert sorted(ds.directory_paths) == sorted(expected_dir_paths)
    assert sorted(ds.file_paths) == sorted(expected_file_paths)

    # total archive sizes should make sense
    assert ds.compressed_size_bytes > 0
    assert ds.compressed_size_bytes < 1e8
    assert ds.uncompressed_size_bytes > 0
    assert ds.uncompressed_size_bytes > ds.compressed_size_bytes

    # size_by_file_extension should work as expected
    assert ds.size_by_file_extension == {
        ".cfg": 258,
        ".md": 15,
        "no-extension": 382,
        ".py": 70,
        ".txt": 11603,
    }

    # files_by_extension makes sense
    assert ds.files_by_extension.keys() == ds.size_by_file_extension.keys()
    assert (
        sum(len(file_list) for file_list in ds.files_by_extension.values())
        == ds.num_files
    )

    # size_by_file_extension should return results sorted from largest to smallest by file size
    last_size_seen = float("inf")
    for size_in_bytes in ds.size_by_file_extension.values():
        assert size_in_bytes < last_size_seen
        last_size_seen = size_in_bytes

    # get_largest_files() should return a non-empty list of _FileInfo objects
    largest_files = ds.get_largest_files(n=2)
    assert isinstance(largest_files, list)
    assert len(largest_files) == 2
    assert all(isinstance(x, _FileInfo) for x in largest_files)
    assert all(x in ds.files for x in largest_files)

    # should actually choose the 2 largest files
    assert largest_files[0].name == "base-package-0.1.0/LICENSE.txt"
    assert largest_files[1].name == "base-package-0.1.0/setup.cfg"


def test_distribution_summary_get_largest_files_works():
    ds = _DistributionSummary.from_file(
        os.path.join(TEST_DATA_DIR, BASE_PACKAGE_SDISTS[0])
    )

    # get_largest_files() should return a non-empty list of _FileInfo objects
    num_files = ds.num_files

    # if you ask for more files than there are, you get however many there are
    assert len(ds.get_largest_files(n=num_files + 1)) == num_files

    # if you ask for less files than there are, you get exactly that many
    assert len(ds.get_largest_files(n=num_files - 1)) == num_files - 1


@pytest.mark.parametrize("distro_file", BASE_WHEELS)
def test_distribution_summary_correctly_reads_contents_of_wheels(distro_file):
    ds = _DistributionSummary.from_file(os.path.join(TEST_DATA_DIR, distro_file))

    # should correctly capture the contents:
    #   * 4 directories
    #   * 8 files
    assert len(ds.files) == 8
    assert ds.num_files == 8

    # `pip wheel` seems to omit directory members from the archive,
    # and `auditwheel repair` seems to restore them
    #
    # so this project's test macOS wheels don't have directory members
    if "macosx" in distro_file:
        expected_dir_paths = []
        expected_file_format = "Mach-O"
        expected_num_directories = 0
        shared_lib_ext = "dylib"
    else:
        expected_dir_paths = [
            "baseballmetrics-0.1.0.dist-info/",
            "baseballmetrics.libs/",
            "baseballmetrics/",
            "lib/",
        ]
        expected_file_format = "ELF"
        expected_num_directories = 4
        shared_lib_ext = "so"

    assert len(ds.directories) == expected_num_directories
    assert ds.num_directories == expected_num_directories

    expected_file_paths = [
        "baseballmetrics-0.1.0.dist-info/METADATA",
        "baseballmetrics-0.1.0.dist-info/RECORD",
        "baseballmetrics-0.1.0.dist-info/WHEEL",
        "baseballmetrics-0.1.0.dist-info/entry_points.txt",
        "baseballmetrics/__init__.py",
        "baseballmetrics/_shared_lib.py",
        "baseballmetrics/metrics.py",
        f"lib/lib_baseballmetrics.{shared_lib_ext}",
    ]

    expected_all_paths = expected_dir_paths + expected_file_paths

    assert sorted(ds.all_paths) == sorted(expected_all_paths)

    # pulling this _FileInfo object out and comparing attribute-by-attribute,
    # to avoid needing to change the test when the library size changes
    assert len(ds.compiled_objects) == 1
    lib_file_info = ds.compiled_objects[0]
    assert lib_file_info.name == f"lib/lib_baseballmetrics.{shared_lib_ext}"
    assert lib_file_info.file_format == expected_file_format
    assert lib_file_info.file_extension == f".{shared_lib_ext}"
    assert lib_file_info.is_compiled is True

    assert sorted(ds.directory_paths) == sorted(expected_dir_paths)
    assert sorted(ds.file_paths) == sorted(expected_file_paths)

    # total archive sizes should make sense
    assert ds.compressed_size_bytes > 0
    assert ds.uncompressed_size_bytes > 0
    if "macosx" in distro_file:
        assert ds.compressed_size_bytes < 25e3
        assert ds.uncompressed_size_bytes != ds.compressed_size_bytes
    else:
        assert ds.compressed_size_bytes < 6e3
        assert ds.uncompressed_size_bytes > ds.compressed_size_bytes

    # size_by_file_extension should work as expected
    if "macosx" in distro_file:
        assert ds.size_by_file_extension == {
            ".dylib": 16504,
            "no-extension": 1129,
            ".py": 536,
            ".txt": 0,
        }
    else:
        assert ds.size_by_file_extension == {
            ".so": 14248,
            "no-extension": 1204,
            ".py": 536,
            ".txt": 0,
        }

    # size_by_file_extension should return results sorted from largest to smallest by file size
    last_size_seen = float("inf")
    for size_in_bytes in ds.size_by_file_extension.values():
        assert size_in_bytes < last_size_seen
        last_size_seen = size_in_bytes


@pytest.mark.parametrize("distro_file", [*BASE_WHEELS, *WINDOWS_WHEELS])
def test_file_info_correctly_determines_file_type(distro_file):
    full_path = os.path.join(TEST_DATA_DIR, distro_file)
    if "macosx" in distro_file:
        expected_file_format = "Mach-O"
        shared_lib_file = "lib/lib_baseballmetrics.dylib"
    elif "manylinux" in distro_file:
        expected_file_format = "ELF"
        shared_lib_file = "lib/lib_baseballmetrics.so"
    else:
        expected_file_format = "Windows PE"
        shared_lib_file = "lightgbm/lib_lightgbm.dll"
    with zipfile.ZipFile(full_path, mode="r") as zf:
        zip_info = zf.getinfo(shared_lib_file)
        file_info = _FileInfo.from_zipfile_member(archive_file=zf, zip_info=zip_info)
        assert file_info.is_compiled is True
        assert file_info.file_format == expected_file_format



import os
import pathlib
import tarfile
import zipfile
from dataclasses import dataclass
from typing import Union

from ._compat import _import_zstandard


@dataclass
class _DirectoryInfo:
    name: str


class _ArchiveFormat:
    BZIP2_TAR = ".tar.bz2"
    CONDA = ".conda"
    GZIP_TAR = ".tar.gz"
    ZIP = ".zip"


def _guess_archive_format(filename: str) -> str:
    if filename.lower().endswith("gz"):
        return _ArchiveFormat.GZIP_TAR
    if filename.lower().endswith("bz2"):
        return _ArchiveFormat.BZIP2_TAR
    if filename.lower().endswith(".conda"):
        return _ArchiveFormat.CONDA
    if filename.lower().endswith(".zip"):
        return _ArchiveFormat.ZIP
    if filename.lower().endswith(".whl"):
        return _ArchiveFormat.ZIP

    raise ValueError(
        f"File '{filename}' does not appear to be a Python package distribution in "
        "one of the formats supported by 'pydistcheck'. "
        "Supported formats: .conda, .tar.bz2, .tar.gz, .whl, .zip"
    )


@dataclass
class _FileInfo:
    name: str
    file_format: str
    file_extension: str
    is_compiled: bool
    uncompressed_size_bytes: int

    @classmethod
    def from_tarfile_member(
        cls, *, archive_file: tarfile.TarFile, tar_info: tarfile.TarInfo
    ) -> "_FileInfo":
        member_name = tar_info.name
        file_format, is_compiled = _guess_archive_member_file_format(
            archive_file=archive_file, member_name=member_name
        )
        return cls(
            name=member_name,
            file_format=file_format,
            file_extension=pathlib.Path(member_name).suffix or "no-extension",
            is_compiled=is_compiled,
            uncompressed_size_bytes=tar_info.size,
        )

    @classmethod
    def from_zipfile_member(
        cls, *, archive_file: zipfile.ZipFile, zip_info: zipfile.ZipInfo
    ) -> "_FileInfo":
        member_name = zip_info.filename
        file_format, is_compiled = _guess_archive_member_file_format(
            archive_file=archive_file, member_name=member_name
        )
        return cls(
            name=member_name,
            file_format=file_format,
            file_extension=pathlib.Path(member_name).suffix or "no-extension",
            is_compiled=is_compiled,
            uncompressed_size_bytes=zip_info.file_size,
        )


# references:
#   * https://en.wikipedia.org/wiki/List_of_file_signatures
#   * https://github.com/apple-oss-distributions/xnu/blob/5c2921b07a2480ab43ec66f5b9e41cb872bc554f/EXTERNAL_HEADERS/mach-o/loader.h#L65
#   * https://github.com/apple-oss-distributions/cctools/blob/658da8c66b4e184458f9c810deca9f6428a773a5/include/mach-o/fat.h#L48
#   * https://github.com/matthew-brett/delocate/blob/df86dddd7c94a93b5c03948b8c127ba0777e2a4d/delocate/tools.py#L166
#   * https://learn.microsoft.com/en-us/previous-versions/ms809762(v=msdn.10)
#   * https://en.wikipedia.org/wiki/Portable_Executable

# DOS MZ (.dll, .exe)
_DOS_MZ_MAGIC_FIRST_2_BYTES = {
    0x5A4D.to_bytes(2, "little"),
}

# ELF (.so, .o)
_ELF_MAGIC_FIRST_4_BYTES = {
    0x7F454C46.to_bytes(4, "little"),
    0x7F454C46.to_bytes(4, "big"),
}

# MACH-O (.dylib)
_MACH_O_MAGIC_FIRST_4_BYTES = {
    0xFEEDFACE.to_bytes(4, "little"),  # 32-bit
    0xFEEDFACE.to_bytes(4, "big"),  # 32-bit
    0xFEEDFACF.to_bytes(4, "little"),  # 64-bit
    0xFEEDFACF.to_bytes(4, "big"),  # 64-bit
    0xCAFEBABE.to_bytes(4, "big"),  # mach-o fat binary
    0xCAFEBABF.to_bytes(4, "big"),  # mach-o fat binary
}


class _FileFormat:
    ELF = "ELF"
    MACH_O = "Mach-O"
    OTHER = "Other"
    WINDOWS_PE = "Windows PE"


def _guess_archive_member_file_format(
    *, archive_file: Union[tarfile.TarFile, zipfile.ZipFile], member_name: str
) -> tuple[str, bool]:
    """
    The approach in this function was inspired by similar code in
    https://github.com/matthew-brett/delocate, so that project's license is included
    in distributions of ``pydistcheck`` as file ``DELOCATE_LICENSE``.

    Returns a two-item tuple of the form ``(file_format, is_compiled)``.
    """
    if isinstance(archive_file, zipfile.ZipFile):
        with archive_file.open(name=member_name, mode="r") as f:
            header = f.read(4)
    else:
        fileobj = archive_file.extractfile(member_name)
        if fileobj is None:  # pragma: no cover
            error_msg = (
                f"'{member_name}' not found. This is a bug in pydistcheck."
                "Report it at https://github.com/jameslamb/pydistcheck/issues."
            )
            raise RuntimeError(error_msg)
        header = fileobj.read(4)

    if header in _ELF_MAGIC_FIRST_4_BYTES:
        return _FileFormat.ELF, True
    if header in _MACH_O_MAGIC_FIRST_4_BYTES:
        return _FileFormat.MACH_O, True
    if header[:2] in _DOS_MZ_MAGIC_FIRST_2_BYTES:
        return _FileFormat.WINDOWS_PE, True

    return _FileFormat.OTHER, False


def _decompress_zstd_archive(*, tar_zst_file: str, decompressed_tar_path: str) -> None:
    """Given a path to a .tar.zst file, decompress its contents to a .tar file"""
    zstandard = _import_zstandard()
    with open(tar_zst_file, "rb") as compressed:
        decompressor = zstandard.ZstdDecompressor()
        with open(decompressed_tar_path, "wb") as destination:
            decompressor.copy_stream(compressed, destination)


def _extract_subset_of_files_from_archive(
    *, archive_file: str, archive_format: str, relative_paths: list[str], out_dir: str
) -> None:
    """
    Extract a subset of files from an archive to a destination directory.

    Extracts AT LEAST those files... might in some cases extract all files.
    """
    if archive_format == _ArchiveFormat.ZIP:
        with zipfile.ZipFile(archive_file, mode="r") as zf:
            zf.extractall(path=out_dir, members=relative_paths)
    elif archive_format == _ArchiveFormat.BZIP2_TAR:
        with tarfile.open(archive_file, mode="r:bz2") as tf:
            tf.extractall(
                path=out_dir,
                members=[tf.getmember(p) for p in relative_paths],
                filter="data",
            )
    elif archive_format == _ArchiveFormat.GZIP_TAR:
        with tarfile.open(archive_file, mode="r:gz") as tf:
            tf.extractall(
                path=out_dir,
                members=[tf.getmember(p) for p in relative_paths],
                filter="data",
            )
    elif archive_format == _ArchiveFormat.CONDA:
        inner_tar_zst_files = []
        # extract any files at the outer ZIP level,
        # and pull out the .tar.zst files
        with zipfile.ZipFile(archive_file, mode="r") as zf:
            for zip_info in zf.infolist():
                # case 1 - skip directories
                if zip_info.is_dir():
                    continue  # pragma: no cover

                # case 2 - if file is in the outer ZIP archive, extract it
                if not zip_info.filename.endswith("tar.zst"):  # pragma: no cover
                    if zip_info.filename in relative_paths:
                        zf.extractall(path=out_dir, members=[zip_info.filename])
                    continue

                # case 3 - one of the zstandard-compressed archives
                inner_tar_zst_files.append(os.path.join(out_dir, zip_info.filename))
                zf.extractall(path=out_dir, members=[zip_info.filename])

        # extract any of these files found in those .tar.zst files
        for tar_zst_file in inner_tar_zst_files:
            # decompress the .tar.zst to just .tar
            decompressed_tar_path = tar_zst_file.replace(".tar.zst", ".tar")
            _decompress_zstd_archive(
                tar_zst_file=tar_zst_file, decompressed_tar_path=decompressed_tar_path
            )

            # do tarfile things
            with tarfile.open(decompressed_tar_path, mode="r") as tf:
                files_to_extract = [
                    tar_info
                    for tar_info in tf.getmembers()
                    if tar_info.name in relative_paths
                ]
                tf.extractall(
                    path=out_dir,
                    members=files_to_extract,
                    filter="data",
                )



"""
Code that prints diagnostic information about a distribution.
"""

from typing import TYPE_CHECKING

from ._utils import _FileSize

if TYPE_CHECKING:
    from ._config import _Config
    from .distribution_summary import _DistributionSummary


def inspect_distribution(*, summary: "_DistributionSummary", config: "_Config") -> None:
    print("file size")
    unit_str = config.output_file_size_unit
    compressed_size = _FileSize(
        num=summary.compressed_size_bytes,
        unit_str="B",
    )
    compressed_size_str = compressed_size.to_string(
        precision=config.output_file_size_precision,
        unit_str=unit_str,
    )
    uncompressed_size = _FileSize(
        num=summary.uncompressed_size_bytes,
        unit_str="B",
    )
    uncompressed_size_str = uncompressed_size.to_string(
        precision=config.output_file_size_precision,
        unit_str=unit_str,
    )
    print(f"  * compressed size: {compressed_size_str}")
    print(f"  * uncompressed size: {uncompressed_size_str}")
    space_saving = 1.0 - (
        compressed_size.total_size_bytes / uncompressed_size.total_size_bytes
    )
    print(f"  * compression space saving: {round(100 * space_saving, 1)}%")

    print("contents")
    print(f"  * directories: {summary.num_directories}")
    print(f"  * files: {summary.num_files} ({len(summary.compiled_objects)} compiled)")

    print("size by extension")
    for extension, size in summary.size_by_file_extension.items():
        size_pct = size / summary.uncompressed_size_bytes
        size_str = _FileSize(
            num=size,
            unit_str="B",
        ).to_string(precision=config.output_file_size_precision, unit_str=unit_str)
        print(f"  * {extension} - {size_str} ({round(size_pct * 100, 1)}%)")

    largest_files = summary.get_largest_files(n=5)
    print("largest files")
    for file_info in largest_files:
        size_str = _FileSize(
            num=file_info.uncompressed_size_bytes,
            unit_str="B",
        ).to_string(precision=config.output_file_size_precision, unit_str=unit_str)
        print(f"  * ({size_str}) {file_info.name}")



"""
miscellaneous helper classes and functions that are
not specific to package distributions
"""

import re

# references:
#
#   * https://physics.nist.gov/cuu/Units/binary.html
#   * https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#meaning-of-memory
#
_UNIT_TO_NUM_BYTES = {
    "b": 1,
    "k": 1024,
    "kb": 1000,
    "ki": 1024,
    "m": 1024**2,
    "mb": 1000000,
    "mi": 1024**2,
    "g": 1024**3,
    "gb": 1000000000,
    "gi": 1024**3,
}


def _recommend_size_str(num_bytes: int) -> tuple[float, str]:
    if num_bytes < int(0.1 * 1024):
        return float(num_bytes), "B"
    if num_bytes <= (0.1 * 1024**2):
        return float(num_bytes) / 1024.0, "K"
    if num_bytes <= (0.1 * 1024**3):
        return float(num_bytes) / (1024**2), "M"

    return float(num_bytes) / (1024**3), "G"


class _FileSize:
    def __init__(
        self,
        *,
        num: float,
        unit_str: str,
    ):
        self._num = num
        self._unit_str = unit_str

    @classmethod
    def from_number(cls, num: int) -> "_FileSize":
        num_bytes, unit_str = _recommend_size_str(num_bytes=num)
        return cls(num=num_bytes, unit_str=unit_str)

    @classmethod
    def from_string(cls, size_str: str) -> "_FileSize":
        parsed = re.search(r"^([0-9\.]+)([A-Za-z]+)$", size_str.strip())
        if parsed is None:
            raise ValueError(f"Could not parse '{size_str}' as a file size.")
        return cls(num=float(parsed.group(1)), unit_str=parsed.group(2))

    def to_string(self, precision: int, unit_str: str) -> str:
        if unit_str == "auto":
            num, unit_str = _recommend_size_str(self.total_size_bytes)
        else:
            num = self.total_size_bytes / _UNIT_TO_NUM_BYTES[unit_str.lower()]
        return f"{round(num, precision)}{unit_str}"

    @property
    def total_size_bytes(self) -> int:
        return int(self._num * _UNIT_TO_NUM_BYTES[self._unit_str.lower()])

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, type(self))
            and self.total_size_bytes == other.total_size_bytes
        )

    def __ge__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes >= other.total_size_bytes

    def __gt__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes > other.total_size_bytes

    def __le__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes <= other.total_size_bytes

    def __lt__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes < other.total_size_bytes

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __str__(self) -> str:
        return self.to_string(precision=3, unit_str="auto")



import os
import re
from sys import platform
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner, Result

import pydistcheck.cli  # noqa: F401
from pydistcheck.cli import check

BASE_PACKAGES = [
    "base-package-0.1.0.tar.gz",
    "base-package-0.1.0.zip",
]
PROBLEMATIC_PACKAGES = [
    "problematic-package-0.1.0.tar.gz",
    "problematic-package-0.1.0.zip",
]
MACOS_SUFFIX = "macosx_10_15_x86_64.macosx_11_6_x86_64.macosx_12_5_x86_64.whl"
MANYLINUX_SUFFIX = "manylinux_2_28_x86_64.manylinux_2_5_x86_64.manylinux1_x86_64.whl"
BASEBALL_CONDA_PACKAGES = [
    "osx-64-baseballmetrics-0.1.0-0.conda",
    "osx-64-baseballmetrics-0.1.0-0.tar.bz2",
]
BASEBALL_WHEELS = [
    f"baseballmetrics-0.1.0-py3-none-{MACOS_SUFFIX}",
    f"baseballmetrics-0.1.0-py3-none-{MANYLINUX_SUFFIX}",
]
BASEBALL_PACKAGES = BASEBALL_CONDA_PACKAGES + BASEBALL_WHEELS
# NOTE: .bz2 and .tar.gz packages here are just unzipped
#       and re-tarred Python wheels... to avoid pydistcheck
#       implicitly assuming that, for example, '*.tar.gz' must
#       be an sdist and therefore not have compiled objects
PACKAGES_WITH_DEBUG_SYMBOLS = [
    "debug-baseballmetrics-0.1.0-macosx-wheel.tar.bz2",
    "debug-baseballmetrics-0.1.0-macosx-wheel.tar.gz",
    f"debug-baseballmetrics-0.1.0-py3-none-{MACOS_SUFFIX}",
    f"debug-baseballmetrics-py3-none-{MANYLINUX_SUFFIX}",
    "osx-64-debug-baseballmetrics-0.1.0-0.conda",
    "osx-64-debug-baseballmetrics-0.1.0-0.tar.bz2",
]
TEST_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

# see https://github.com/jameslamb/pydistcheck/issues/206
NUMPY_WIN_DEBUG_WHL = "numpy-1.26.3-cp310-cp310-win_amd64.whl"
if os.path.isfile(os.path.join(TEST_DATA_DIR, NUMPY_WIN_DEBUG_WHL)):
    PACKAGES_WITH_DEBUG_SYMBOLS.append(NUMPY_WIN_DEBUG_WHL)


def _assert_log_matches_pattern(
    result: Result, pattern: str, num_times: int = 1
) -> None:
    log_lines = result.output.split("\n")
    match_results = [bool(re.search(pattern, log_line)) for log_line in log_lines]
    num_matches_found = sum(match_results)
    msg = (
        f"Expected to find 1 instance of '{pattern}' in logs, "
        f"found {num_matches_found} in {log_lines}."
    )
    assert num_matches_found == num_times, msg


def _mock_check_num_calls(mock_check: MagicMock) -> int:
    """returns count of times the __call__() method of a mocked check class instance was called"""
    return sum(
        1
        for mc in mock_check.method_calls
        if str(mc).startswith("call()(distro_summary=")
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES + BASEBALL_PACKAGES)
def test_check_runs_without_error(distro_file):
    result = CliRunner().invoke(check, [os.path.join(TEST_DATA_DIR, distro_file)])
    assert result.exit_code == 0


def test_version_flag_works():
    result = CliRunner().invoke(
        check,
        [
            "--version",
        ],
    )
    assert result.exit_code == 0

    _assert_log_matches_pattern(
        result=result,
        pattern=(r"^pydistcheck [0-9]{1}\.[0-9]{1,2}\.[0-9]{1,2}\.*[0-9]*$"),
        num_times=1,
    )


def test_check_fails_with_informative_error_if_file_doesnt_exist():
    result = CliRunner().invoke(check, ["some-garbage.exe"])
    # NOTE: this exit code comes from 'click'
    assert result.exit_code >= 1
    _assert_log_matches_pattern(result, r"Path 'some\-garbage\.exe' does not exist\.$")


def test_check_runs_for_all_files_before_exiting():
    result = CliRunner().invoke(
        check,
        [
            "--max-allowed-size-compressed=5B",
            os.path.join(TEST_DATA_DIR, "base-package-0.1.0.tar.gz"),
            os.path.join(TEST_DATA_DIR, "base-package-0.1.0.zip"),
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(5\.0B\)\.$"
        ),
        num_times=2,
    )
    _assert_log_matches_pattern(
        result=result, pattern="errors found while checking\\: 1", num_times=2
    )


@pytest.mark.parametrize("flags", ([], ["--inspect"]))
def test_check_fails_with_informative_error_if_file_is_an_unrecognized_format(flags):
    result = CliRunner().invoke(check, [__file__, *flags])
    assert result.exit_code == 2
    _assert_log_matches_pattern(
        result, r"error.*File '.*' does not appear to be a Python package distribution"
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_ignore_with_one_check(distro_file):
    runner = CliRunner()

    # fails when running all checks
    result = runner.invoke(
        check, [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"]
    )
    assert result.exit_code == 1

    # fails with --ignore if the failing check isn't mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=path-contains-spaces",
            "--max-allowed-files=1",
        ],
    )
    assert result.exit_code == 1

    # succeeds if the failing check is mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=too-many-files",
            "--max-allowed-files=1",
        ],
    )
    assert result.exit_code == 0
    _assert_log_matches_pattern(result, "errors found while checking\\: 0")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_ignore_with_multiple_checks(distro_file):
    runner = CliRunner()

    # fails when running all checks
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--max-allowed-files=1",
            "--max-allowed-size-compressed=1B",
        ],
    )
    assert result.exit_code == 1

    # fails with --ignore if all of the failing checks aren't mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=path-contains-spaces",
            "--ignore=too-many-files",
            "--max-allowed-files=1",
            "--max-allowed-size-compressed=1B",
        ],
    )
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(1\.0B\)\.$"
        ),
    )
    assert result.exit_code == 1

    # succeeds if the failing checks are mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=distro-too-large-compressed",
            "--ignore=too-many-files",
            "--max-allowed-files=1",
        ],
    )
    assert result.exit_code == 0
    _assert_log_matches_pattern(result, "errors found while checking\\: 0")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_fails_with_expected_error_if_one_check_in_ignore_is_unrecognized(
    distro_file,
):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=random-nonsense",
            "--ignore=too-many-files",
        ],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"ERROR\: found the following unrecognized checks passed via '\-\-ignore'\: "
            r"random\-nonsense"
        ),
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_fails_with_expected_error_if_multiple_checks_in_ignore_are_unrecognized(
    distro_file,
):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=garbage",
            "--ignore=other-trash",
            "--ignore=random-nonsense",
            "--ignore=too-many-files",
        ],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"ERROR\: found the following unrecognized checks passed via '\-\-ignore'\: "
            r"garbage,other\-trash,random\-nonsense"
        ),
    )


@patch("pydistcheck.cli._SpacesInPathCheck", autospec=True)
@patch("pydistcheck.cli._FileCountCheck", autospec=True)
@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_select_with_one_check(
    mock_path_contains_spaces, mock_too_many_files, distro_file
):
    # ensure this attribute is actually set, as it's used for filtering (the class attribute
    # is lost somewhere in all this mocking and patching)
    mock_path_contains_spaces.return_value.check_name = "path-contains-spaces"
    mock_too_many_files.return_value.check_name = "too-many-files"

    # initialize the click testing class
    runner = CliRunner()

    # by default, all checks should be selected and run once
    result = runner.invoke(check, [os.path.join(TEST_DATA_DIR, distro_file)])
    assert result.exit_code == 0
    assert _mock_check_num_calls(mock_path_contains_spaces) == 1
    assert _mock_check_num_calls(mock_too_many_files) == 1
    mock_path_contains_spaces.reset_mock()
    mock_too_many_files.reset_mock()

    # passing '--select' switches pydistcheck to opt-in mode... no other checks should run
    # except those mentioned in '--select'
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--select=too-many-files",
        ],
    )
    assert result.exit_code == 0
    _assert_log_matches_pattern(result, "errors found while checking\\: 0")
    assert _mock_check_num_calls(mock_path_contains_spaces) == 0
    assert _mock_check_num_calls(mock_too_many_files) == 1
    mock_path_contains_spaces.reset_mock()
    mock_too_many_files.reset_mock()

    # if a check is present in '--select', it doesn't matter that it's present in '--ignore'
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=too-many-files",
            "--select=too-many-files",
        ],
    )
    assert result.exit_code == 0
    assert _mock_check_num_calls(mock_path_contains_spaces) == 0
    assert _mock_check_num_calls(mock_too_many_files) == 1
    mock_path_contains_spaces.reset_mock()
    mock_too_many_files.reset_mock()


@patch("pydistcheck.cli._PathTooLongCheck", autospec=True)
@patch("pydistcheck.cli._SpacesInPathCheck", autospec=True)
@patch("pydistcheck.cli._FileCountCheck", autospec=True)
@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_select_with_multiple_checks(
    mock_path_too_long, mock_path_contains_spaces, mock_too_many_files, distro_file
):
    # ensure this attribute is actually set, as it's used for filtering (the class attribute
    # is lost somewhere in all this mocking and patching)
    mock_path_too_long.return_value.check_name = "path-too-long"
    mock_path_contains_spaces.return_value.check_name = "path-contains-spaces"
    mock_too_many_files.return_value.check_name = "too-many-files"

    # initialize the click testing class
    runner = CliRunner()

    # by default, all checks should be selected and run once
    result = runner.invoke(check, [os.path.join(TEST_DATA_DIR, distro_file)])
    assert result.exit_code == 0
    assert _mock_check_num_calls(mock_path_too_long) == 1
    assert _mock_check_num_calls(mock_path_contains_spaces) == 1
    assert _mock_check_num_calls(mock_too_many_files) == 1
    mock_path_too_long.reset_mock()
    mock_path_contains_spaces.reset_mock()
    mock_too_many_files.reset_mock()

    # passing '--select' switches pydistcheck to opt-in mode... no other checks should run
    # except those mentioned in '--select'
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--select=too-many-files",
            "--select=path-too-long",
        ],
    )
    assert result.exit_code == 0
    _assert_log_matches_pattern(result, "errors found while checking\\: 0")
    assert _mock_check_num_calls(mock_path_too_long) == 1
    assert _mock_check_num_calls(mock_path_contains_spaces) == 0
    assert _mock_check_num_calls(mock_too_many_files) == 1
    mock_path_too_long.reset_mock()
    mock_path_contains_spaces.reset_mock()
    mock_too_many_files.reset_mock()


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_fails_with_expected_error_if_one_check_in_select_is_unrecognized(
    distro_file,
):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--select=random-nonsense",
            "--ignore=too-many-files",
        ],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"ERROR\: found the following unrecognized checks passed via '\-\-select'\: "
            r"random\-nonsense"
        ),
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_fails_with_expected_error_if_multiple_checks_in_select_are_unrecognized(
    distro_file,
):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--select=garbage",
            "--select=other-trash",
            "--select=random-nonsense",
            "--select=too-many-files",
        ],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"ERROR\: found the following unrecognized checks passed via '\-\-select'\: "
            r"garbage,other\-trash,random\-nonsense"
        ),
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_max_allowed_files(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check, [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"]
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result, r"^1\. \[too\-many\-files\] Found 11 files\. Only 1 allowed\.$"
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize(
    "size_str",
    [
        "1K",
        "3.999K",
        "0.005KB",
        "0.0002M",
        "0.00000008G",
        "708B",
    ],
)
@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_max_allowed_size_compressed(size_str, distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            f"--max-allowed-size-compressed={size_str}",
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \([0-9]+\.[0-9]+[BKMG]\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("unit_str", ["B", "K", "KB", "Mi", "GB", "Gi"])
def test_check_error_messages_supports_output_file_size_unit(unit_str):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, BASE_PACKAGES[1]),
            "--max-allowed-size-compressed=1B",
            "--max-allowed-size-uncompressed=1B",
            f"--output-file-size-unit={unit_str}",
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            rf"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+{unit_str} is "
            rf"larger than the allowed size \([0-9]+\.[0-9]+{unit_str}\)\.$"
        ),
    )
    _assert_log_matches_pattern(
        result,
        (
            rf"^2\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+{unit_str} is "
            rf"larger than the allowed size \([0-9]+\.[0-9]+{unit_str}\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 2")


def test_check_error_messages_supports_output_file_size_precision():
    # default precision
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, BASE_PACKAGES[1]),
            "--max-allowed-size-compressed=256B",
            "--max-allowed-size-uncompressed=256B",
            "--output-file-size-unit=MB",
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size 0\.007MB is "
            r"larger than the allowed size \(0\.0MB\)\.$"
        ),
    )
    _assert_log_matches_pattern(
        result,
        (
            r"^2\. \[distro\-too\-large\-uncompressed\] Uncompressed size 0\.012MB is "
            r"larger than the allowed size \(0\.0MB\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 2")

    # precision 5
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, BASE_PACKAGES[1]),
            "--max-allowed-size-compressed=256B",
            "--max-allowed-size-uncompressed=256B",
            "--output-file-size-precision=5",
            "--output-file-size-unit=MB",
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size 0\.00681MB is "
            r"larger than the allowed size \(0\.00026MB\)\.$"
        ),
    )
    _assert_log_matches_pattern(
        result,
        (
            r"^2\. \[distro\-too\-large\-uncompressed\] Uncompressed size 0\.01233MB is "
            r"larger than the allowed size \(0\.00026MB\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 2")


@pytest.mark.parametrize(
    "size_str",
    [
        "1",
        "K",
        "1-KB",
        "GB1",
        "1G-B",
    ],
)
@pytest.mark.parametrize(
    "cli_arg",
    [
        "--max-allowed-size-compressed",
        "--max-allowed-size-uncompressed",
    ],
)
def test_check_raises_informative_error_for_malformed_file_size_config(
    cli_arg, size_str
):
    with pytest.raises(
        ValueError, match=rf"Could not parse '{size_str}' as a file size"
    ):
        CliRunner().invoke(
            check,
            [
                os.path.join(TEST_DATA_DIR, BASE_PACKAGES[0]),
                f"{cli_arg}={size_str}",
            ],
            catch_exceptions=False,
        )


@pytest.mark.parametrize(
    "size_str",
    [
        "1K",
        "3.999K",
        "0.005KB",
        "0.0002M",
        "0.00000008G",
        "708B",
    ],
)
@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_max_allowed_size_uncompressed(size_str, distro_file):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            f"--max-allowed-size-uncompressed={size_str}",
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \([0-9]+\.[0-9]+[BKMG]\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_prefers_pyproject_toml_to_defaults(distro_file, tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with open("pyproject.toml", "w") as f:
            f.write(
                "[tool.pylint]\n[tool.pydistcheck]\nmax_allowed_size_uncompressed = '7B'\n"
            )
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file)],
        )

    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(7\.0B\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_handles_ignore_list_in_pyproject_toml_correctly(distro_file, tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"],
        )
        assert result.exit_code == 1
        _assert_log_matches_pattern(
            result, r"^1\. \[too\-many\-files\] Found [0-9]+ files\. Only 1 allowed\.$"
        )
        with open("pyproject.toml", "w") as f:
            f.write(
                "[tool.pylint]\n[tool.pydistcheck]\n"
                "ignore = [\n'too-many-files', 'path-contains-non-ascii-characters'\n]\n"
            )
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"],
        )
        assert result.exit_code == 0
        _assert_log_matches_pattern(result, "errors found while checking\\: 0")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_prefers_keyword_args_to_pyproject_toml_and_defaults(
    distro_file, tmp_path
):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with open("pyproject.toml", "w") as f:
            f.write(
                "[tool.pylint]\n[tool.pydistcheck]\nmax_allowed_size_uncompressed = '7B'\n"
            )
        result = runner.invoke(
            check,
            [
                os.path.join(TEST_DATA_DIR, distro_file),
                "--max-allowed-size-uncompressed=123B",
            ],
        )

    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(0\.12K\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_prefers_custom_toml_file_to_root_pyproject_toml(distro_file, tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with open("pyproject.toml", "w") as f:
            f.write(
                "[tool.pylint]\n[tool.pydistcheck]\nmax_allowed_size_uncompressed = '10GB'\n"
            )
        other_dir = os.path.join(tmp_path, "cool-files")
        other_config = os.path.join(other_dir, "stuff.toml")
        os.mkdir(other_dir)
        with open(other_config, "w") as f:
            f.write(
                "[tool.pylint]\n[tool.pydistcheck]\nmax_allowed_size_uncompressed = '7B'\n"
            )
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file), f"--config={other_config}"],
        )

    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(7\.0B\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_files_only_differ_by_case_works(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^1\. \[files\-only\-differ\-by\-case\] Found files which differ only by case\. "
            r"Files\: problematic\-package\-0\.1\.0/problematic_package/Question\.py"
            r",problematic\-package\-0\.1\.0/problematic_package/question\.PY"
            r",problematic\-package\-0\.1\.0/problematic_package/question\.py"
        ),
    )
    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]{1}"
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_mixed_file_extension_use_works(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^2\. \[mixed\-file\-extensions\] Found a mix of file extensions for the "
            r"same file type\: \.NDJSON \(1\), \.jsonl \(1\), \.ndjson \(1\)"
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^3\. \[mixed\-file\-extensions\] Found a mix of file extensions for the "
            r"same file type\: \.yaml \(2\), \.yml \(1\)"
        ),
    )

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]{1}"
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_path_contains_non_ascii_characters_works(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    # NOTE: matching a variable number of '?' because zip and tar (and even different
    #       implementations of those) encode non-ASCII characters in paths differently
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^4\. \[path\-contains\-non\-ascii\-characters\] Found file path containing non-ASCII "
            r"characters\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/\?+veryone-loves-python\.py'"
        ),
    )

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]{1}"
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_path_contains_spaces_works(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    # should report a file with spaces in a directory without spaces
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^5\. \[path\-contains\-spaces\] "
            r"Found path with spaces\: 'problematic\-package\-0\.1\.0/beep boop\.ini"
        ),
    )

    # should report a directory with spaces
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^6\. \[path\-contains\-spaces\] Found path with spaces\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/bad code[/]*"
        ),
    )

    # should report a file without spaces inside a directory with spaces
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^7\. \[path\-contains\-spaces\] Found path with spaces\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/bad code/__init__\.py"
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^8\. \[path\-contains\-spaces\] Found path with spaces\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/bad code/ship\-it\.py"
        ),
    )

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]{1}"
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_unexpected_files_check_works(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    # directory
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^9\. \[unexpected\-files\] Found unexpected directory "
            r"'problematic\-package\-0\.1\.0/\.git[/]{0,1}'\."
        ),
    )

    # root-level files
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^10\. \[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/\.gitignore'\."
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^11\. \[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/\.hadolint\.yaml'\."
        ),
    )

    # nested files
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^12\. \[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/problematic_package/\.gitignore'\."
        ),
    )

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]{1}"
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_unexpected_files_correctly_respects_multiple_cli_args(distro_file):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--expected-files=!*.gitignore",
            "--expected-files=!*.git/HEAD",
            "--expected-directories=src/",
        ],
    )
    assert result.exit_code == 1

    # directories: should report that the expected directory was not found
    _assert_log_matches_pattern(
        result=result,
        pattern=r"\[expected\-files\] Did not find any directories matching pattern 'src/'",
    )

    # files: should find BOTH .gitignore files
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"\[unexpected\-files\] Found unexpected file 'problematic\-package\-0\.1\.0/\.gitignore"
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"\[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/problematic_package/\.gitignore"
        ),
    )

    # files: should find the one .git/HEAD file
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"\[unexpected\-files\] Found unexpected file 'problematic\-package\-0\.1\.0/\.git/HEAD"
        ),
    )

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]{1}"
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_expected_files_check_works(distro_file):
    result = CliRunner().invoke(
        check,
        [
            "--expected-files=*.R",
            "--expected-directories=Movies/*",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 1

    # directory
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^1\. \[expected\-files\] Did not find any directories matching pattern 'Movies/\*'."
        ),
    )

    # file
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^2\. \[expected\-files\] Did not find any files matching pattern '\*\.R'."
        ),
    )

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: [0-9]+"
    )


@pytest.mark.parametrize("distro_file", BASEBALL_PACKAGES)
def test_expected_files_does_not_raise_check_failure_if_all_patterns_match(distro_file):
    result = CliRunner().invoke(
        check,
        [
            # (wheel) lib/lib_baseball_metrics
            # (conda) lib/python3.9/site-packages/lib/lib_baseballmetrics.dylib
            "--expected-files=*/lib_baseballmetrics.*",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 0

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: 0"
    )


def test_expected_files_does_not_raise_check_failure_if_directory_pattern_matches():
    # conda packages, macOS wheels do not preserve directory members...
    # testing with a manylinux wheel to test that directory functionality works
    distro_file = f"baseballmetrics-0.1.0-py3-none-{MANYLINUX_SUFFIX}"
    result = CliRunner().invoke(
        check,
        [
            "--expected-directories=lib/",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 0

    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: 0"
    )


@pytest.mark.parametrize("distro_file", BASEBALL_CONDA_PACKAGES)
def test_path_too_long_check_works_for_conda_packages(distro_file):
    result = CliRunner().invoke(
        check,
        [
            "--max-path-length=5",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"\[path\-too\-long\] Path too long \(78 > 5\)\: "
            r"'lib/python3\.9/site\-packages/baseballmetrics/__pycache__/metrics\.cpython\-39\.pyc'"
        ),
    )


@pytest.mark.parametrize("distro_file", BASEBALL_WHEELS)
def test_path_too_long_check_works_for_wheels(distro_file):
    result = CliRunner().invoke(
        check,
        [
            "--max-path-length=5",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"\[path\-too\-long\] Path too long \(30 > 5\)\: "
            r"'baseballmetrics/_shared_lib\.py'"
        ),
    )


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_cli_raises_exactly_the_expected_number_of_errors_for_the_problematic_package(
    distro_file,
):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result=result, pattern=r"errors found while checking\: 12$"
    )


@pytest.mark.parametrize("distro_file", PACKAGES_WITH_DEBUG_SYMBOLS)
def test_debug_symbols_check_works(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1, result.output
    if "macosx" in distro_file:
        # macOS wheels
        lib_file = r"\"lib/lib_baseballmetrics\.dylib\"'"
        if platform.startswith(("cygwin", "win")):
            debug_cmd = r"'llvm\-nm \-a " + lib_file
        else:
            # dsymutil works on both macOS and Linux
            debug_cmd = r"'dsymutil \-s " + lib_file
    elif "osx-64" in distro_file:
        # macOS conda packages
        lib_file = r"\"lib/python3\.9/site-packages/lib/lib_baseballmetrics\.dylib\"'\."
        if platform.startswith(("cygwin", "win")):
            debug_cmd = r"'llvm\-nm \-a " + lib_file
        else:
            # dsymutil works on both macOS and Linux
            debug_cmd = r"'dsymutil \-s " + lib_file
    elif NUMPY_WIN_DEBUG_WHL in distro_file:
        # windows wheels
        lib_file = r"\"numpy\.libs/libopenblas64__v0\.3\.23-293-gc2f4bdbb-gcc_10_3_0-2bde3a66a51006b2b53eb373ff767a3f\.dll\"'"
        debug_cmd = r"'objdump \-\-all\-headers " + lib_file
    else:
        # linux wheels
        debug_cmd = r"'objdump \-\-all\-headers \"lib/lib_baseballmetrics\.so\"'\."

    expected_msg = (
        r"^1\. \[compiled\-objects\-have\-debug\-symbols\] "
        r"Found compiled object containing debug symbols\. "
        r"For details, extract the distribution contents and run "
    )
    expected_msg += debug_cmd
    _assert_log_matches_pattern(result=result, pattern=expected_msg)


# --------------------- #
# pydistcheck --inspect #
# --------------------- #


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_inspect_runs_before_checks(distro_file):
    result = CliRunner().invoke(
        check,
        [
            "--inspect",
            "--max-allowed-files=1",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result, r"^1\. \[too\-many\-files\] Found 11 files\. Only 1 allowed\.$"
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")
    _assert_log_matches_pattern(result, r"^size by extension$")


def test_inspect_respects_output_file_size_unit_for_all_size_strings():
    distro_file = os.path.join(TEST_DATA_DIR, BASE_PACKAGES[0])

    # --output-file-size-unit auto
    result = CliRunner().invoke(
        check,
        [
            "--inspect",
            distro_file,
        ],
    )
    assert result.exit_code == 0

    # mix of B and K
    _assert_log_matches_pattern(result, r" compressed size: 4\.938K")
    _assert_log_matches_pattern(result, r" uncompressed size: 12\.039K")
    _assert_log_matches_pattern(result, r" \.py \- 70\.0B")
    _assert_log_matches_pattern(
        result, r" \(11\.092K\) base-package\-0\.1\.0/LICENSE\.txt"
    )

    # --output-file-size-unit B
    result = CliRunner().invoke(
        check,
        [
            "--inspect",
            "--output-file-size-unit=B",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 0

    # all B
    _assert_log_matches_pattern(result, r" compressed size: 5057\.0B")
    _assert_log_matches_pattern(result, r" uncompressed size: 12328\.0B")
    _assert_log_matches_pattern(result, r" \.py \- 70\.0B")
    _assert_log_matches_pattern(
        result, r" \(11358\.0B\) base-package\-0\.1\.0/LICENSE\.txt"
    )


def test_inspect_respects_output_file_size_precision_for_all_size_strings():
    distro_file = os.path.join(TEST_DATA_DIR, BASE_PACKAGES[0])

    # --output-file-size-unit auto
    result = CliRunner().invoke(
        check,
        [
            "--inspect",
            distro_file,
        ],
    )
    assert result.exit_code == 0

    # default precision
    _assert_log_matches_pattern(result, r" compressed size: 4\.938K")
    _assert_log_matches_pattern(result, r" uncompressed size: 12\.039K")
    _assert_log_matches_pattern(result, r" \.py \- 70\.0B")
    _assert_log_matches_pattern(
        result, r" \(11\.092K\) base-package\-0\.1\.0/LICENSE\.txt"
    )

    # --output-file-size-precision 2
    result = CliRunner().invoke(
        check,
        [
            "--inspect",
            "--output-file-size-precision=2",
            os.path.join(TEST_DATA_DIR, distro_file),
        ],
    )
    assert result.exit_code == 0

    # precision 2
    _assert_log_matches_pattern(result, r" compressed size: 4\.94K")
    _assert_log_matches_pattern(result, r" uncompressed size: 12\.04K")
    _assert_log_matches_pattern(result, r" \.py \- 70\.0B")
    _assert_log_matches_pattern(
        result, r" \(11\.09K\) base-package\-0\.1\.0/LICENSE\.txt"
    )



"""
Implementations for individual checks that ``pydistcheck``
performs on distributions.
"""

import os
from collections import defaultdict
from collections.abc import Sequence
from fnmatch import fnmatchcase
from tempfile import TemporaryDirectory
from typing import Protocol

from ._distribution_summary import _DistributionSummary
from ._file_utils import _extract_subset_of_files_from_archive
from ._shared_lib_utils import _file_has_debug_symbols
from ._utils import _FileSize

# ALL_CHECKS constant is used to validate configuration options like '--ignore' that reference
# check names. It's a set literal so it doesn't need to be recomputed at runtime, and this project
# relies on unit tests to ensure that it's updated as the list of checks changes.
ALL_CHECKS = {
    "compiled-objects-have-debug-symbols",
    "distro-too-large-compressed",
    "distro-too-large-uncompressed",
    "expected-files",
    "too-many-files",
    "files-only-differ-by-case",
    "mixed-file-extensions",
    "path-contains-non-ascii-characters",
    "path-contains-spaces",
    "path-too-long",
    "unexpected-files",
}


class _CheckProtocol(Protocol):
    check_name: str

    def __call__(
        self, distro_summary: _DistributionSummary
    ) -> list[str]:  # pragma: no cover
        ...


class _CompiledObjectsDebugSymbolCheck(_CheckProtocol):
    check_name = "compiled-objects-have-debug-symbols"

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        compiled_object_paths = [
            file_info.name for file_info in distro_summary.compiled_objects
        ]
        if not compiled_object_paths:
            return out

        with TemporaryDirectory() as tmp_dir:
            _extract_subset_of_files_from_archive(
                archive_file=distro_summary.original_file,
                archive_format=distro_summary.archive_format,
                relative_paths=compiled_object_paths,
                out_dir=tmp_dir,
            )

            for file_relative_path in compiled_object_paths:
                has_debug_symbols, cmd_str = _file_has_debug_symbols(
                    file_absolute_path=os.path.join(tmp_dir, file_relative_path)
                )
                if has_debug_symbols:
                    msg = (
                        f"[{self.check_name}] Found compiled object containing debug symbols. "
                        "For details, extract the distribution contents and run "
                        f"'{cmd_str} \"{file_relative_path}\"'."
                    )
                    out.append(msg)
        return out


class _DistroTooLargeCompressedCheck(_CheckProtocol):
    check_name = "distro-too-large-compressed"

    def __init__(
        self,
        *,
        max_allowed_size_bytes: int,
        output_file_size_precision: int,
        output_file_size_unit: str,
    ):
        self.max_allowed_size_bytes = max_allowed_size_bytes
        self.output_file_size_precision = output_file_size_precision
        self.output_file_size_unit = output_file_size_unit

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        max_size = _FileSize(num=self.max_allowed_size_bytes, unit_str="B")
        actual_size = _FileSize(num=distro_summary.compressed_size_bytes, unit_str="B")
        if actual_size > max_size:
            actual_size_str = actual_size.to_string(
                precision=self.output_file_size_precision,
                unit_str=self.output_file_size_unit,
            )
            max_size_str = max_size.to_string(
                precision=self.output_file_size_precision,
                unit_str=self.output_file_size_unit,
            )
            msg = (
                f"[{self.check_name}] Compressed size {actual_size_str} is larger "
                f"than the allowed size ({max_size_str})."
            )
            out.append(msg)
        return out


class _DistroTooLargeUnCompressedCheck(_CheckProtocol):
    check_name = "distro-too-large-uncompressed"

    def __init__(
        self,
        *,
        max_allowed_size_bytes: int,
        output_file_size_precision: int,
        output_file_size_unit: str,
    ):
        self.max_allowed_size_bytes = max_allowed_size_bytes
        self.output_file_size_precision = output_file_size_precision
        self.output_file_size_unit = output_file_size_unit

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        max_size = _FileSize(num=self.max_allowed_size_bytes, unit_str="B")
        actual_size = _FileSize(
            num=distro_summary.uncompressed_size_bytes, unit_str="B"
        )
        if actual_size > max_size:
            actual_size_str = actual_size.to_string(
                precision=self.output_file_size_precision,
                unit_str=self.output_file_size_unit,
            )
            max_size_str = max_size.to_string(
                precision=self.output_file_size_precision,
                unit_str=self.output_file_size_unit,
            )
            msg = (
                f"[{self.check_name}] Uncompressed size {actual_size_str} is larger "
                f"than the allowed size ({max_size_str})."
            )
            out.append(msg)
        return out


class _FileCountCheck(_CheckProtocol):
    check_name = "too-many-files"

    def __init__(self, *, max_allowed_files: int):
        self.max_allowed_files = max_allowed_files

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        num_files = distro_summary.num_files
        if num_files > self.max_allowed_files:
            msg = (
                f"[{self.check_name}] Found {num_files} files. "
                f"Only {self.max_allowed_files} allowed."
            )
            out.append(msg)
        return out


class _FilesOnlyDifferByCaseCheck(_CheckProtocol):
    check_name = "files-only-differ-by-case"

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        path_lower_to_raw = defaultdict(list)
        for file_path in distro_summary.all_paths:
            path_lower_to_raw[file_path.lower()].append(file_path)

        duplicates_list: list[str] = []
        for filepaths in path_lower_to_raw.values():
            if len(filepaths) > 1:
                duplicates_list += filepaths

        if duplicates_list:
            duplicates_str = ",".join(sorted(duplicates_list))
            msg = (
                f"[{self.check_name}] Found files which differ only by case. "
                f"Files: {duplicates_str}"
            )
            out.append(msg)
        return out


class _NonAsciiCharacterCheck(_CheckProtocol):
    check_name = "path-contains-non-ascii-characters"

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        for file_path in distro_summary.all_paths:
            if not file_path.isascii():
                ascii_converted_str = file_path.encode("ascii", "replace").decode(
                    "ascii"
                )
                msg = (
                    f"[{self.check_name}] Found file path containing non-ASCII characters: "
                    f"'{ascii_converted_str}'"
                )
                out.append(msg)
        return out


class _MixedFileExtensionCheck(_CheckProtocol):
    check_name = "mixed-file-extensions"

    file_ext_groups = (
        {".cc", ".CC", ".cpp", ".CPP"},
        {".htm", ".HTM", ".html", ".HTML"},
        {".jpg", ".JPG", ".jpeg", ".JPEG"},
        {".jsonl", ".JSONL", ".ndjson", ".NDJSON"},
        {".txt", ".TXT", ".text", ".TEXT"},
        {".yaml", ".YAML", ".yml", ".YML"},
    )

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        file_extensions_in_distro = set(distro_summary.files_by_extension.keys())
        for file_ext_group in self.file_ext_groups:
            extensions_found = file_ext_group.intersection(file_extensions_in_distro)
            if len(extensions_found) >= 2:
                count_str = ", ".join(
                    f"{ext} ({distro_summary.count_by_file_extension[ext]})"
                    for ext in sorted(extensions_found)
                )
                msg = (
                    f"[{self.check_name}] Found a mix of file extensions for "
                    f"the same file type: {count_str}"
                )
                out.append(msg)
        return out


class _PathTooLongCheck(_CheckProtocol):
    check_name = "path-too-long"

    def __init__(self, *, max_path_length: int):
        self.max_path_length = max_path_length

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        bad_paths = [
            p for p in distro_summary.all_paths if len(p) > self.max_path_length
        ]
        for file_path in bad_paths:
            msg = (
                f"[{self.check_name}] Path too long ({len(file_path)} > {self.max_path_length}): "
                f"'{file_path}'"
            )
            out.append(msg)
        return out


class _SpacesInPathCheck(_CheckProtocol):
    check_name = "path-contains-spaces"

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        for file_path in distro_summary.all_paths:
            if file_path != file_path.replace(" ", ""):
                msg = f"[{self.check_name}] Found path with spaces: '{file_path}'"
                out.append(msg)
        return out


class _ExpectedFilesCheck(_CheckProtocol):
    check_name = "expected-files"

    def __init__(
        self,
        *,
        directory_patterns: Sequence[str],
        file_patterns: Sequence[str],
    ):
        self.directory_patterns = [
            d for d in directory_patterns if not d.startswith("!")
        ]
        self.file_patterns = [f for f in file_patterns if not f.startswith("!")]

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        for pattern in self.file_patterns:
            found_any = False
            for file_path in distro_summary.file_paths:
                if fnmatchcase(file_path, pattern):
                    found_any = True
                    break

            if not found_any:
                msg = f"[{self.check_name}] Did not find any files matching pattern '{pattern}'."
                out.append(msg)

        for pattern in self.directory_patterns:
            found_any = False
            for directory_path in distro_summary.directory_paths:
                # NOTE: some archive formats have a trailing "/" on directory names,
                #       some do not
                if fnmatchcase(directory_path, pattern) or fnmatchcase(
                    directory_path, pattern + "/"
                ):
                    found_any = True
                    break
            if not found_any:
                msg = f"[{self.check_name}] Did not find any directories matching pattern '{pattern}'."
                out.append(msg)
        return out


class _UnexpectedFilesCheck(_CheckProtocol):
    check_name = "unexpected-files"

    def __init__(
        self,
        *,
        directory_patterns: Sequence[str],
        file_patterns: Sequence[str],
    ):
        self.directory_patterns = [
            d[1:] for d in directory_patterns if d.startswith("!")
        ]
        self.file_patterns = [f[1:] for f in file_patterns if f.startswith("!")]

    def __call__(self, distro_summary: _DistributionSummary) -> list[str]:
        out: list[str] = []
        for file_path in distro_summary.file_paths:
            for pattern in self.file_patterns:
                if fnmatchcase(file_path, pattern):
                    msg = f"[{self.check_name}] Found unexpected file '{file_path}'."
                    out.append(msg)

        for directory_path in distro_summary.directory_paths:
            for pattern in self.directory_patterns:
                # NOTE: some archive formats have a trailing "/" on directory names,
                #       some do not
                if fnmatchcase(directory_path, pattern) or fnmatchcase(
                    directory_path, pattern + "/"
                ):
                    msg = f"[{self.check_name}] Found unexpected directory '{directory_path}'."
                    out.append(msg)
        return out



"""
Tests that ensure documentation accurately describes the
state of the project's source code.
"""

from pathlib import Path

from pydistcheck._checks import ALL_CHECKS
from pydistcheck._compat import tomllib
from pydistcheck._config import _ALLOWED_CONFIG_VALUES, _Config

DOCS_ROOT = Path(__file__).parents[1].joinpath("docs")


def test_default_toml_config():
    # NOTE: this intentionally does not use pydistcheck._Config.update_from_toml(),
    #       to ensure that bugs in that class don't also cause this test to accidentally pass.
    defaults_example_file = DOCS_ROOT.joinpath("_static/defaults.toml")
    with open(defaults_example_file, "rb") as f:
        config_dict = tomllib.load(f)
        tool_options = config_dict.get("tool", {}).get("pydistcheck", {})

    opts_in_docs = set(tool_options.keys())
    missing_opts = _ALLOWED_CONFIG_VALUES - opts_in_docs
    assert len(missing_opts) == 0, f"missing options: {','.join(missing_opts)}"
    extra_opts = opts_in_docs - _ALLOWED_CONFIG_VALUES
    assert len(extra_opts) == 0, f"unrecognized options: {','.join(extra_opts)}"

    # values should exactly match the defaults
    config = _Config().update_from_toml(defaults_example_file)

    # The default in _Config class is a tuple because 'dataclasses' requires something immutable,
    # but it's read in as a list from pyproject.toml.
    #
    # This allows for that deviation.
    config.expected_directories = tuple(config.expected_directories)
    config.expected_files = tuple(config.expected_files)
    config.ignore = ()
    config.select = ()

    assert config == _Config(), (
        "values in 'docs/_static/defaults.toml' do not match actual defaults used by pydistcheck"
    )


def test_all_checks_are_documented_in_check_reference():
    check_ref_file = DOCS_ROOT.joinpath("check-reference.rst")
    with open(check_ref_file) as f:
        check_ref_str = f.read()

    for check in ALL_CHECKS:
        assert f"\n\n{check}\n****" in check_ref_str, (
            f"'{check}' not yet documented in 'docs/check-reference.rst'"
        )



import sys
from unittest import mock

import pytest

from pydistcheck._compat import _import_zstandard


def test_import_zstandard_raises_informative_error_if_it_isnt_found():
    with mock.patch.dict(sys.modules):
        sys.modules["zstandard"] = None

        with pytest.raises(
            ModuleNotFoundError,
            match="Checking zstd-compressed files requires the 'zstandard' library",
        ):
            _import_zstandard()



import subprocess
from unittest.mock import Mock, patch

from pydistcheck._shared_lib_utils import (
    _MACHO_STRIP_SYMBOL,
    _get_symbols,
    _run_command,
)


def test_run_command_handles_binary_output():
    """Test that _run_command can handle binary output containing invalid UTF-8 bytes."""
    mock_output = bytes(
        [0x68, 0x65, 0x6C, 0x6C, 0x6F, 0xD7, 0x77, 0x6F, 0x72, 0x6C, 0x64]
    )  # i.e., "hello" + invalid UTF-8 + "world"

    with patch("subprocess.run") as mock_run:
        mock_run.return_value = Mock(stdout=mock_output)

        # Should not raise UnicodeDecodeError,
        result = _run_command(["some", "command"])
        # latin1 encoding should preserve all bytes.
        assert len(result) == len(mock_output)
        # Check valid parts of the string.
        assert "hello" in result
        assert "world" in result


def test_run_command_handles_command_failure():
    with patch("subprocess.run") as mock_run:
        mock_run.side_effect = subprocess.CalledProcessError(1, ["cmd"])
        result = _run_command(["failing", "command"])
        assert result == "__command_failed__"


def test_run_command_handles_missing_command():
    with patch("subprocess.run") as mock_run:
        mock_run.side_effect = FileNotFoundError()
        result = _run_command(["nonexistent"])
        assert result == "__tool_not_available__"


def test_get_symbols_filters_radr_symbol():
    """Test that _get_symbols filters out the Mach-O strip symbol."""
    mock_output = (
        "0000000000000000 T _main\n"
        f"0000000005614542 - 00 0000    OPT {_MACHO_STRIP_SYMBOL}\n"
        "0000000000001000 D _data\n"
    )

    with patch("subprocess.run") as mock_run:
        mock_run.return_value = Mock(stdout=mock_output.encode())

        result = _get_symbols(["nm", "-a"], "test.dylib")

        # Should keep real symbols but filter out the strip symbol.
        assert "_main" in result
        assert "_data" in result
        assert _MACHO_STRIP_SYMBOL not in result
        assert len(result.split("\n")) == 2  # Only two real symbols
